@@ -22,18 +22,20 @@
 #define LUV_THREAD_MAXNUM_ARG 9
 
 typedef struct {
-  /* only support LUA_TNIL, LUA_TBOOLEAN, LUA_TLIGHTUSERDATA, LUA_TNUMBER, LUA_TSTRING*/
+  // support basic lua type LUA_TNIL, LUA_TBOOLEAN, LUA_TNUMBER, LUA_TSTRING
+  // and support uv_handle_t userdata
   int type;
   union
   {
     lua_Number num;
     int boolean;
-    void* userdata;
+    void* userdata;   // luv private scope uv_handle_t
     struct {
       const char* base;
       size_t len;
     } str;
   } val;
+  int ref;            // ref of uv_handle_t
 } luv_val_t;
 
 typedef struct {
@@ -41,8 +43,18 @@ typedef struct {
   luv_val_t argv[LUV_THREAD_MAXNUM_ARG];
 } luv_thread_arg_t;
 
-static int luv_thread_arg_set(lua_State* L, luv_thread_arg_t* args, int idx, int top, int flag);
-static int luv_thread_arg_push(lua_State* L, const luv_thread_arg_t* args);
-static void luv_thread_arg_clear(luv_thread_arg_t* args);
+//luajit miss LUA_OK
+#ifndef LUA_OK
+#define LUA_OK 0
+#endif
+
+//LUV flags thread support userdata handle
+#define LUVF_THREAD_UHANDLE 1
+
+#ifdef LUV_SOURCE
+static int luv_thread_arg_set(lua_State* L, luv_thread_arg_t* args, int idx, int top, int flags);
+static int luv_thread_arg_push(lua_State* L, luv_thread_arg_t* args, int flags);
+static void luv_thread_arg_clear(lua_State* L, luv_thread_arg_t* args, int flags);
+#endif
 
 #endif //LUV_LTHREADPOOL_H
